/*

Program Name:	TTT.CPP
Author(s):      Written by Turrell Taylor, Paul Travis, and Antony Gordon
Class:          POS/370 - Introduction to Programming in C++
Instructor:     Craig Kaerscher
Meeting Time:   Tuesday, 6 pm

Last modified:  04 FEB 2003
$Revision: 1.5 $

Description:

	This program demonstrates the culmination of our studies
	in and outside of the classroom. This program will model
	the human thought process in a simple game of Tic Tac Toe.
	The human opponent is 'O' and the computer 'X', which can
	be changed in the code.

	Note: This code as written will only compile in Borland C/C++
	compilers. Any attempt to compile in MS/Visual C/C++ will cause
	undue programming stress.

Analyst		Date				Description
-------------------------------------------------------------------------
adg		   02/04      		Added two player support
									Modified single player game to randomly choose
									which player goes first. Updated to Release
									Candidate 2
adg			01/30				Made code compliant with C and C++ compilers
									Cleaned up displayboard () code
adg			01/29				Cleaned up source code more, added separate
									procedure for displaying score board
adg			01/28				Fixed core logic bug in canwin()
adg			01/27				Cleaned up source code to make it easier to follow
adg			01/20				Added UI enhancements to make display nicer,
									foregoing any ability to compile in Visual C++
									without too much headache
adg			01/17				Initial install
*/

#include <time.h>
#include <stdio.h>
#include <conio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

/* Define our default constants */

#ifndef FALSE
#define FALSE   0
#endif
#ifndef TRUE
#define TRUE   !FALSE
#endif


/*    Global variable section


		The Tic Tac Toe board is a 3x3 table. We will represent it in
		C as a 3x3 2 dimensional array to make the logic easy to follow.

		The main instance used in the program will be square, which is globally
		accessible.

*/


/*
		I automatically increment BUILDNO when something is added/changed
*/

const int BUILDNO = 137;

const char COMPUTER = 'X';
const char OPPONENT = 'O';
const char EMPXY = '*';
const char PROGNAME[12] = "Tic Tac Toe";
const char PROGREL [34] = "Release Candidate 3 Build %04d";
#ifdef __cplusplus
const int MAXSIZE = 3;
const int STARTBOARDX = 33;
const int STARTBOARDY = 9;
#else
#define MAXSIZE 3
#define STARTBOARDX 33
#define STARTBOARDY 9
#endif

/*
	ANSI C does not allow const int types to be used with arrays. C++ allows this
	so I use compiler directives to make sure that the code works either way.

	I prefer const int rather than using #define.
*/

typedef char BOARD [MAXSIZE][MAXSIZE];

/*
	Global variables
*/
char VERSION[35];
int win;
int moves;
int o_score;
int c_score;
int scratch_count;
int game_count;
int twoplayerflag;
BOARD square;

/*

		  Function declarations

*/

int checkwin (char);
int canwin (char);
void initboard ();
void getmove ();
int getnumplayers ();
void displayboard ();
void makemove ();
void scoreboard ();
void singleplayer ();
void twoplayer ();
void center (int, const char *);
/*

  Begin main execution

*/
int main()

/*
        main () is the entry point of the program
        All global variables are initialized and program
        control is transferred based on the users choice
        of single or two player.
*/


{
        clrscr ();
	srand((unsigned int)time(NULL)); /* Initialize the random number generator */
	o_score=c_score=scratch_count=0; /* one massive initialize!!! */
	game_count = 1;
	if (getnumplayers())
	  twoplayer ();  /* Play the two player version */
	else
	  singleplayer (); /* Play the single player version */
return 0;

}

void center (int row, const char *s)
/*
	This function centers a string constant
*/

{
  int x;

  x = ( (80 - strlen (s)) / 2 );
  gotoxy (x, row);
  puts(s);
}

int canwin ( char XorO )
/*

  canwin detects any possibility of the computer winning
  and attempts to prevent the player from winning

*/

{
  int row, col;

  row=0;   /* Start at the top row and ... */
  do {
	 col=0;  /* the top column */
	 do {
               /* Go through the 'board' and look for an empty spot */
		if (square [row][col] == EMPXY) {
		  square [row][col] = XorO; /* Place a character there */
		  if (!checkwin (XorO)) /* See if there is a win */
			 square [row][col] = EMPXY; /* if not, clear it out */
		  else {
                  /* We can't let the player easily win, so if we're checking
                     on them, and we find a potential...*/
			 if (XorO == OPPONENT) {
				square [row][col] = COMPUTER; /* we block it */
				return TRUE;
			 }
			 return TRUE;
		  }
		}
		col++;
	 } while (col < MAXSIZE);
	 row++;
  } while (row<MAXSIZE);
  return FALSE;
}

int checkwin ( char ltr )

/*
	 This function scans the 'board' for possible win scenarios

	 ltr is set to letter we want to check
	 checkwin returns TRUE if a win is detected.

	 uses square [][]
*/

{
  int ct;

  /* First check all rows and columns */

  for (ct=0; ct < MAXSIZE; ct++ ) {
	 if ((square [ct][0] == ltr) && (square [ct][1] == ltr) && (square [ct][2] == ltr))
		return TRUE;
	 if ((square [0][ct] == ltr) && (square [1][ct] == ltr) && (square [2][ct] == ltr))
		return TRUE;
  }

  /* Check the diagonals */

  if ((square [0][0] == ltr) && (square [1][1] == ltr) && (square [2][2] == ltr))
	 return TRUE;
  if ((square [0][2] == ltr) && (square [1][1] == ltr) && (square [2][0] == ltr))
	 return TRUE;

/* Obviously we don't have a winner so... */

	 return FALSE;
}

void displayboard ()

{
  int x, y, r, c;

  scoreboard ();
  for (x=STARTBOARDX, c=0; x<46; x+=(2*MAXSIZE),c++)
	 for (y=STARTBOARDY,r=0; y<14; y+=2,r++) {
		gotoxy (x, y);
		printf ("%c", square [r][c]);
		if (x <= (STARTBOARDX+6))
		  printf("  |");
	 }
  gotoxy (STARTBOARDX-2, STARTBOARDY+1);
  printf ("-----+-----+-----\n");
  gotoxy (STARTBOARDX-2, STARTBOARDY+3);
  printf ("-----+-----+-----\n");

  gotoxy (STARTBOARDX, STARTBOARDY-2);
  printf ("1     2     3");
  for (r=1,c=STARTBOARDY; r<8; r+=3,c+=2) {
	 gotoxy (STARTBOARDX-6, c);
	 printf ("%d", r);
  }
}

void getmove (char player)
/*

	getmove () accepts a single digit and converts it into
	a row,col pair for use with the 2 dimensional array

*/
{
  int good, loc, row, col;
  char c[2];

  good = FALSE; /* Default to a bad move */
  while (!good) {
	 do {
		gotoxy (8, 19); clreol ();
		printf ("Enter location from 1 to 9 to place %c in: ", player);
                if (toupper(c[0] == 'Q'))
                  exit(1);
		loc=atoi(fgets(c, sizeof(c), stdin));
		fflush (stdin); /* Clears the buffer for the next go round */
	 } while ((loc < 1) || (loc >(MAXSIZE * MAXSIZE)));
	 loc--;  /* Make loc zero based so that it will correctly move through the table */
	 row = loc / MAXSIZE; col = loc % MAXSIZE;
	 if ( square [row][col] != EMPXY ) {
		displayboard ();
		gotoxy (8, 19); clreol ();
		printf("Invalid move, try again...");
		while (!kbhit()) ;
	 }
	 else {
		square [row][col] = player;
		good = TRUE;
	 }
  }
}

int getnumplayers ()
/*
	getnumplayers () asks how many people are playing, and then
	and returns a boolean value directing program flow into the
	single or two player version
*/
{
	  char c[2];
        printf(PROGNAME);
	printf("\nHow many players 1,2 [Q=quit]: ");
	fgets(c, sizeof(c), stdin);
	fflush (stdin);
	if (toupper(c[0])=='Q')
	  exit (0);
	else
	  return (atoi(c) == 2);
}

void initboard ()

/*
	initboard () initializes the board and randomly places the computer's
	character in the table. This random placement insures that the game is
	not predictable

*/
{

  int row, col;

  for (row=0; row < MAXSIZE; row++)
	 for (col=0; col < MAXSIZE; col++)
		square [row][col] = EMPXY;
}

void makemove ()

/*
	makemove () evaluates the board and sees where the computer can make a
	move.

*/

{
  if ( square[0][0] == EMPXY )
	 square[0][0] = COMPUTER;
	 else if ( square [0][MAXSIZE-1] == EMPXY )
		square [0][2] = COMPUTER;
		else if ( square [MAXSIZE-1][0] == EMPXY )
		  square [MAXSIZE-1][0] = COMPUTER;
		  else if ( square [MAXSIZE-1][MAXSIZE-1] == EMPXY )
			 square [MAXSIZE-1][MAXSIZE-1] = COMPUTER;
			 else if ( square [0][1] == EMPXY )
				square [0][1] = COMPUTER;
				else if ( square [1][0] == EMPXY )
				  square [1][0] = COMPUTER;
				  else if ( square [1][1] == EMPXY )
					 square [1][1] = COMPUTER;
					 else if ( square [1][MAXSIZE-1] == EMPXY )
						square [1][MAXSIZE-1] = COMPUTER;
						else if (square [MAXSIZE-1][1] == EMPXY )
						  square [MAXSIZE-1][1] = COMPUTER;
  win = checkwin (COMPUTER);
}

void scoreboard ()
/* Displays the scores... */
{
  char SCOREBOARD[50];
  #ifdef DEBUG
	 sprintf (SCOREBOARD, "%c Score %d\t%c Score %d\tDraw %d\tMoves %d",OPPONENT, o_score,
	 COMPUTER, c_score, scratch_count, moves);
  #else
	 sprintf (SCOREBOARD, "%c Score %d\t%c Score %d\tDraw %d",OPPONENT, o_score,
	 COMPUTER, c_score, scratch_count);
  #endif
  gotoxy(18, 16);
  puts(SCOREBOARD);
}

void singleplayer ()
/*

	This procedure handles the single player version of the game.
	The computer always starts first

*/

{

  sprintf(VERSION, PROGREL, BUILDNO);
  do {
	 moves=0;
	 clrscr ();
	 center (1, PROGNAME);
	 center (2, VERSION);
	 center (3, "Single Player Version");
	 initboard ();
	 if (rand() % 2) {
		/* I could have used ?: but this block is a bit more complex */
		square [ rand() % 3 ][ rand() % 3]=COMPUTER; /* Place the computer in a random square */
		displayboard ();
		moves = 2;
	 }
	 else {
		displayboard ();
		getmove (OPPONENT);
		displayboard ();
		makemove ();
		moves = 2;
	 }
	 win = 0;
	 do {
		displayboard ();
		getmove (OPPONENT);
		moves++;
		if ( checkwin (OPPONENT)) {
		  displayboard ();
		  win++;
		  gotoxy (8, 20);
		  printf ("Congratulations! %c won.\n", OPPONENT);
		  o_score++;
		  break;
		}
		else {
		  if (canwin(COMPUTER)) {
			 win++;
			 gotoxy (8, 20);
			 printf ("Sorry, I win.\n");
			 c_score++;
			 break;
		  }
		  else
			 if ( !canwin (OPPONENT))
				makemove ();
			 moves++;
		  if (moves >= 9) {
			 displayboard ();
			 gotoxy (8, 20);
			 printf ("Good game. It's a draw\n");
			 scratch_count++;
			 break;
		  }
		}

	 } while ( (moves < 9) || ( win < 1) );
	 displayboard ();
	 gotoxy (8, 21);
	 printf ("Do you want to play again: (Y/N)");
	 game_count++;
  } while ( toupper (getch()) != 'N' );
}

void twoplayer ()
/*

	This procedure handles the two player version of the game.
	The computer randomly picks who will start first

*/

{
  char ch;

  sprintf(VERSION, PROGREL, BUILDNO);
  do {
	 moves=0;
	 clrscr ();
	 twoplayerflag = TRUE;  /* We are playing a two player game */
	 center (1, PROGNAME);
	 center (2, VERSION);
	 center (3, "Two Player Version");
	 initboard ();
	 ch = ( rand () % 2) ? OPPONENT : COMPUTER; /* This allows random choice in player */
	 displayboard ();
	 getmove (ch);
	 displayboard();
	 if (ch == OPPONENT)
	  getmove (COMPUTER);
	 win = 0;
	 moves=2;
	 do {
		displayboard ();
		getmove (OPPONENT);
		displayboard();
		moves++;
		if ( checkwin (OPPONENT)) {
		  displayboard ();
		  win++;
		  gotoxy (8, 20);
		  printf ("Congratulations! %c won.\n", OPPONENT);
		  o_score++;
		  break;
		}
                getmove (COMPUTER);
		moves++;
		if (checkwin(COMPUTER)) {
		  displayboard ();
		  win++;
		  gotoxy (8, 20);
		  printf ("Congratulations! %c won.\n", COMPUTER);
		  c_score++;
		  break;
		}
		if (moves >= 9) {
		  displayboard ();
		  gotoxy (8, 20);
		  printf ("Good game. It's a draw\n");
		  scratch_count++;
		  break;
		}
		displayboard();
	 } while ( (moves < 9) || ( !win) );
	 displayboard ();
	 gotoxy (8, 21);
	 printf ("Do you want to play again: (Y/N)");
	 game_count++;
  } while ( toupper (getch()) != 'N' );
}
